import java.util.LinkedList;
import java.util.ListIterator;
import java.util.List;

/**
 *  The CircularLLIterator class provides an iterator through
 *  a linear LinkedList object that makes it appear to be
 *  circular.  The element after the last element in the list,
 *  as returned by <code>next</code>, is the first element in
 *  the list.  The element before the first element in the list,
 *  as returned by <code>previous</code>, is the last element
 *  in the list.
 *
 *  <p>
 *  In a circular iteration, the methods <code>hasNext</code>
 *  and <code>hasPrevious</code> will always return
 *  <code>true</code> unless the list is empty.  This means
 *  that the usual iteration pattern of
 *  <pre>
 *     while (iterator.hasNext())
 *         <em>&lt;do something&gt;</em></pre>
 *  will result in an infinite loop.  To iterate once through
 *  the elements of the list, use the list iterator returned by
 *  the linked list's <code>listIterator</code> method or get
 *  the size of the list and iterate through that many times.
 *  (The latter technique will not work if you are adding or
 *  removing elements as you iterate.)
 *
 *  @author Alyce Brady
 *  @version 13 October 2002
 *  @see LinkedList
 **/
public class CircularLLIterator implements ListIterator
{
    private List list;
    private ListIterator internalIterator;

    /** Constructs a circular iterator for the specified list.
     *      @param list the linked list through which to iterate
     **/
    public CircularLLIterator(List list, int startIndex)
    {
        this.list = list;
        internalIterator = list.listIterator(startIndex);
    }

    /** Returns <code>true</code> if this list iterator has more elements
     *  when traversing the list in the forward direction. (In other words,
     *  returns <code>true</code> if <code>next</code> would return an
     *  element rather than throwing an exception.)  Since this is a
     *  circular iterator, returns <code>true</code> unless the list is
     *  empty.
     *      @return <code>true</code> if the list iterator has more elements
     *              when traversing the list in the forward direction
     **/
    public boolean hasNext()
    {
        return list.size() != 0;
    }

    /** Returns <code>true</code> if this list iterator has more elements
     *  when traversing the list in the reverse direction. (In other words,
     *  returns <code>true</code> if <code>previous</code> would return an
     *  element rather than throwing an exception.)  Since this is a
     *  circular iterator, returns <code>true</code> unless the list is
     *  empty.
     *      @return <code>true</code> if the list iterator has more elements
     *              when traversing the list in the reverse direction
     **/
    public boolean hasPrevious()
    {
        return list.size() != 0;
    }

    /** Returns the index of the element that would be returned by a
     *  subsequent call to <code>next</code>. (Returns list size if the
     *  list is empty.)
     **/
    public int nextIndex()
    {
        if ( internalIterator.hasNext() )
            return internalIterator.nextIndex();
        else
            return 0;
    }

    /** Returns the index of the element that would be returned by a
     *  subsequent call to <code>previous</code>. (Returns -1 if the
     *  list is empty.)
     **/
    public int previousIndex()
    {
        if ( list.size() == 0 )
            return -1;

        if ( internalIterator.hasPrevious() )
            return internalIterator.previousIndex();
        else
            return list.size() - 1;
    }

    /** Returns the next element in the list. This method may be called
     *  repeatedly to iterate through the list, or intermixed with calls
     *  to <code>previous</code> to go back and forth. (Note that
     *  alternating calls to <code>next</code> and <code>previous</code>
     *  will return the same element repeatedly.)
     *      @return the next element in the list
     *      @throws java.util.NoSuchElementException if the list is empty
     **/
    public Object next()
    {
        if ( list.size() == 0 )
            return null;

        if ( internalIterator.hasNext() )
            return internalIterator.next();
        else
        {
            internalIterator = list.listIterator(0);
            return internalIterator.next();
        }
    }

    /** Returns the previous element in the list. This method may be called
     *  repeatedly to iterate through the list backwards, or intermixed with
     *  calls to <code>next</code> to go back and forth. (Note that
     *  alternating calls to <code>next</code> and <code>previous</code>
     *  will return the same element repeatedly.)
     *      @return the previous element in the list
     *      @throws java.util.NoSuchElementException if the list is empty
     **/
    public Object previous()
    {
        if ( list.size() == 0 )
            return null;

        if ( internalIterator.hasPrevious() )
            return internalIterator.previous();
        else
        {
            internalIterator = list.listIterator(list.size());
            Object obj = internalIterator.previous();
            return obj;
        }
    }

    /** Inserts the specified element into the list. The element is inserted
     *  immediately before the next element that would be returned by
     *  <code>next</code>, if any, and after the next element that would be
     *  returned by <code>previous</code>, if any. (If the list contains no
     *  elements, the new element becomes the sole element on the list.)
     *  The new element is inserted before the implicit cursor: a subsequent
     *  call to <code>next</code> would be unaffected, and a subsequent call
     *  to <code>previous</code> would return the new element. (This call
     *  increases by one the value that would be returned by a call to
     *  <code>nextIndex</code> or <code>previousIndex</code>.)
     *      @param obj the element to insert
     **/
    public void add(Object obj)
    {
        internalIterator.add(obj);
    }

    /** Removes from the list the last element that was returned by
     *  <code>next</code> or <code>previous</code>. This call can only
     *  be made once per call to <code>next</code> or <code>previous</code>.
     *  It can be made only if <code>ListIterator.add</code> has not been
     *  called after the last call to <code>next</code> or
     *  <code>previous</code>.
     *      @throws IllegalStateException neither <code>next</code> nor
     *              <code>previous</code> have been called, or
     *              <code>remove</code> or <code>add</code> have been called
     *              after the last call to <code>next</code> or
     *              <code>previous</code>
     **/
    public void remove()
    {
        internalIterator.remove();
    }

    /** Replaces the last element returned by <code>next</code> or
     *  <code>previous</code> with the specified element. This call
     *  can be made only if neither <code>ListIterator.remove</code>
     *  nor <code>ListIterator.add</code> have been called after the
     *  last call to <code>next</code> or <code>previous</code>.
     *      @param obj the element with which to replace the last element
     *                 returned by <code>next</code> or <code>previous</code>
     **/
    public void set(Object obj)
    {
        internalIterator.set(obj);
    }
}